  #include <avr/io.h>
  #include <avr/delay.h>
  #include <avr/interrupt.h>
  #include <avr/signal.h>
  #include <compat/twi.h>
  #include <stdio.h>
  #include <string.h>
  
  #define LCD_DDR		DDRB
  #define LCD_PORT 	PORTB
  #define LCD_RS 		0
  #define LCD_E 		1
  #define LCD_DB4 	2
  #define LCD_DB5 	3
  #define LCD_DB6 	4
  #define LCD_DB7 	5
  #define LCD_LED 	6
  #define LCD_VO 		7
  
  #define KEY16_LINE_DDR 			DDRD
  #define KEY16_LINE_PORT 		PORTD
  #define KEY16_LINE_IN 			PIND
  #define KEY16_COLUMN_DDR 		DDRC
  #define KEY16_COLUMN_PORT		PORTC
  #define KEY16_COLUMN_IN 		PINC
  #define KEY16_LINE_1 				4
  #define KEY16_LINE_2 				5
  #define KEY16_LINE_3 				6
  #define KEY16_LINE_4 				7
  #define KEY16_COLUMN_1 			4
  #define KEY16_COLUMN_2 			5
  #define KEY16_COLUMN_3 			6
  #define KEY16_COLUMN_4 			7
  
  #define PS2_DDR 			DDRD
  #define PS2_PORT 			PORTD
  #define PS2_IN 				PIND
  #define PC_DATA 			0
  #define PC_CLOCK 			1
  #define AT_KEY_DATA 	3
  #define AT_KEY_CLOCK 	2
  
  #define AT_KEY_2_PC_SWITCH_DDR		DDRC
  #define AT_KEY_2_PC_SWITCH_PORT		PORTC
  #define AT_KEY_2_PC_SWITCH				2
  
  #define JOY_DDR 			DDRA
  #define JOY_PORT 			PORTA
  #define JOY_IN 				PINA
  #define JOY_BUTTON_0 	0
  #define JOY_BUTTON_1 	1
  #define JOY_BUTTON_2 	2
  #define JOY_BUTTON_3 	3
  #define JOY_POTI_0 		4
  #define JOY_POTI_1 		5
  #define JOY_POTI_2 		6
  #define JOY_POTI_3 		7
  
  #define EXT_EEPROM_DDR 	DDRC
  #define EXT_EEPROM_SCL 	0
  #define EXT_EEPROM_SDA 	1
  
  #define KEY_KEYSET 		13
  #define KEY_JOYSET 		14
  #define KEY_PASSWORD 	15
  #define KEY_SETUP 		16
  #define KEY_OK 				4
  #define KEY_CANCEL 		12
  
  #define TIMER2PRESCALER 		8
  #define DEBOUNCE_TIME 			20E-3
  #define REPEAT_START_TIME 	500E-3
  #define REPEAT_NEXT_TIME 		200E-3
  #define MS2WAIT							500
  
  #define SCL_CLOCK  					100000
  #define EXT_EEPROM_READ    	1
  #define EXT_EEPROM_WRITE   	0
  #define EXT_EEPROM_ADDRESS 	0xA0
  
  #define NO 				0
  #define OK 				1
  #define CANCEL 		2
  
  #define SET_KEY				0
  #define SET_JOY				1
  #define SET_PASSWORD 	2
  
  #define MAX_KEY_DATA 				201
  #define JOYSET_ADDRESS			0x9D08
  #define PASSWORD_ADDRESS 		0x4E84
  #define PIN_ADDRESS 				0xFB40
  #define BACKLIGHT_ADDRESS 	0xFB46
  #define CONTRAST_ADDRESS 		0xFB48
  #define JOYCAL_ADDRESS			0xFB4A
  
  #define SETUP_PROGRAM 		0
  #define SETUP_JOYSTICK 		1
  #define SETUP_PIN					2
  #define SETUP_BACKLIGHT		3
  #define SETUP_CONTRAST 		4
  #define LAST_SETUP_ITEM		4
  
  struct lcdview
  {
    unsigned char count,trig;
  };
  
  struct lcdelement
  {
    struct lcdview led,vo;
  };
  
  struct at_keyboard
  {
    unsigned char clockedge,bitcount,finish,data;
  };
  
  struct a16_keyboard
  {
    unsigned char key,oldkey,prevkey,longpress,repeat;
  };
  
  struct joystick_calibration
  {
    unsigned char trig_mid_max,trig_mid_min;
  };
  
  struct joystick
  {
    unsigned char button[4],oldbuttons,prevbuttons,btn_repeat,btn_longpress,
										poti[4],potiaxis,oldpotiaxis,prevpotiaxis,
										poti_repeat,poti_longpress,potipointer;
    struct joystick_calibration cal[4];
  };
  
  struct buffertype
  {
    unsigned char pointer,data[MAX_KEY_DATA];
  };
  
  volatile struct lcdelement lcd;
  volatile struct at_keyboard keyat;
  volatile struct a16_keyboard key16;
  volatile struct joystick joy;
  volatile struct buffertype keybuffer;
  
  volatile unsigned char timer0count,timer2count,debounce,
  	                        repeat_start,repeat_next,*ledvo_trig,
														keyset,joyset,passwordset,set_mode;
  
  char lcdstr[17];
  
  signed char digit16[17]={-1,7,4,1,-1,8,5,2,0,9,6,3,-1,-1,-1,-1,-1};
  
  
  /* wartet eine Zeit von ca. (us) s */
  void wait_us (unsigned int us)
  {
  unsigned int a;
  
    us=(us*10)/15;
    for (a=1;a<=us;a++)
    {
      _delay_us(1);
    }
  	
  }  // wait_us
  
  /* wartet eine Zeit von ca. (ms) ms */
  void wait_ms (unsigned int ms)
  {
  unsigned int a;
  
    for (a=1;a<=ms;a++)
    {
      _delay_ms(1);
    }
  	
  }  // wait_ms
  
  /* sendet 4-Bit Daten (rs=1) oder Befehle (rs=0) an die LCD-Anzeige */
  void lcd_hex_4 (unsigned char z, unsigned char rs)
  {
  
    if (rs)
    {
      rs=(1<<LCD_RS);
    }
    LCD_PORT=(z<<LCD_DB4) | (1<<LCD_E) | (rs);
    _delay_us(1);
    LCD_PORT=LCD_PORT & ~(1<<LCD_E);
    _delay_us(1);
    LCD_PORT=LCD_PORT | (1<<LCD_E);
  	
  }  // lcd_hex_4
  
  /* sendet 8-Bit Daten (rs=1) oder Befehle (rs=0) an die LCD-Anzeige */
  void lcd_hex_8 (unsigned char z,unsigned char rs)
  {
  
    lcd_hex_4((z>>4) & 0x0F,rs);
    lcd_hex_4(z & 0x0F,rs);
  	
  }  // lcd_hex_8
  
  /* lscht die LCD-Anzeige */
  void lcd_clear (void)
  {
  
    lcd_hex_8(0x01,0);
    _delay_ms(2);
  	
  }  // lcd_clear
  
  /* Initialisiert die LCD-Anzeige */
  void lcd_init (void)
  {
  
    // E High setzen
    LCD_PORT=(1<<LCD_E);
    
    // DB4-DB7, E, RS, LED und VO als Ausgang setzen
    LCD_DDR=(1<<LCD_DB4) | (1<<LCD_DB5) | (1<<LCD_DB6) | (1<<LCD_DB7) | 
            (1<<LCD_RS) | (1<<LCD_E) | (1<<LCD_LED) | (1<<LCD_VO);
    
    wait_ms(40);
    lcd_hex_4(0x3,0);
  
    wait_ms(6);
    lcd_hex_4(0x3,0);
    
    wait_us(150);
    lcd_hex_4(0x3,0);
    wait_us(100);
    
    // 4-Bit Interface setzen
    lcd_hex_4(0x2,0);
    wait_us(50);
  
    // 2 Zeilen LCD-Anzeige und 5x7 Zeichensatz
    lcd_hex_8(0x28,0);  
    wait_us(50);
    
    // LCD-Anzeige an
    lcd_hex_8(0x0C,0);
    wait_us(50);
    
    // LCD-Anzeige lschen
    lcd_clear();
    
    // Entry mode set
    // Adresspointer inkrementieren und Displayinhalt nicht schieben
    lcd_hex_8(0x06,0);
    wait_us(50);
    
    // Initialisierung der Timer-Werte fr Hintergrundbeleuchtung und Kontrast
    lcd.led.count=0;
    lcd.vo.count=0;
  	
  }  // lcd_init
  
  /* Initialisierung der PS/2-Schnittstellen fr Tastatur und PC */
  void ps2_init (void)
  {
  
    // AT_KEY_DATA, AT_KEY_CLOCK, PC_DATA und PC_CLOCK als Eingang setzen
    PS2_DDR=PS2_DDR & ~((1<<AT_KEY_DATA) | (1<<AT_KEY_CLOCK) | 
  			  (1<<PC_DATA) | (1<<PC_CLOCK));
  	
    // Normal-Modus, Tastatur mit PC verbunden
    AT_KEY_2_PC_SWITCH_DDR=AT_KEY_2_PC_SWITCH_DDR | (1<<AT_KEY_2_PC_SWITCH);
    AT_KEY_2_PC_SWITCH_PORT=AT_KEY_2_PC_SWITCH_PORT | (1<<AT_KEY_2_PC_SWITCH);
    
  }  // ps2_init
  
  /* Initialisierung der 16er-Tastatur */
  void key16_init (void)
  {
  
    // Line1-Line4 und Column1-Column4 Pull-Ups
    KEY16_LINE_PORT=KEY16_LINE_PORT | (1<<KEY16_LINE_1) | (1<<KEY16_LINE_2) | 
																			(1<<KEY16_LINE_3) | (1<<KEY16_LINE_4);
    KEY16_COLUMN_PORT=KEY16_COLUMN_PORT | (1<<KEY16_COLUMN_1) | 
																					(1<<KEY16_COLUMN_2) | 
																					(1<<KEY16_COLUMN_3) | 
																					(1<<KEY16_COLUMN_4);
  	
    // Line1-Line4 und Column1-Column4 als Eingang setzen
    KEY16_LINE_DDR=KEY16_LINE_DDR & ~((1<<KEY16_LINE_1) | (1<<KEY16_LINE_2) | 
																			(1<<KEY16_LINE_3) | (1<<KEY16_LINE_4));
    KEY16_COLUMN_DDR=KEY16_COLUMN_DDR & ~((1<<KEY16_COLUMN_1) | 
																					(1<<KEY16_COLUMN_2) | 
																					(1<<KEY16_COLUMN_3) | 
																					(1<<KEY16_COLUMN_4));
    
    // Initialisierung der Variablen
    key16.key=0;
    key16.oldkey=0;
    key16.prevkey=0;
    key16.longpress=0;
    key16.repeat=0;
  	
  }  // key16_init
  
  /* Initialisierung des Joysticks */
  void joy_init (void)
  {
  unsigned char i;
  	
    // Initialisierung der Variablen
    joy.potipointer=0;
    joy.prevbuttons=0;
    joy.oldpotiaxis=0;
    joy.prevpotiaxis=0;
    joy.potiaxis=0;
    joy.btn_longpress=0;
    joy.poti_longpress=0;
    joy.btn_repeat=0;
    joy.poti_repeat=0;
    joy.oldbuttons=0x0F;
    for (i=0;i<=3;i++)
    {
      joy.button[i]=1;
      joy.poti[i]=0;
    }
		
		// PUll-Ups fr Joystick-Knpfe einschalten und alle Pins 
    // als Eingang setzen          
    JOY_PORT=(1<<JOY_BUTTON_0) | (1<<JOY_BUTTON_1) | 
							(1<<JOY_BUTTON_2) | (1<<JOY_BUTTON_3);
    JOY_DDR=0x00;
  	
    // AVcc (5V) als Referenz gewhlt und Ergebnis Linksbndig (8 Bit)
    ADMUX=(1<<REFS0) | (1<<ADLAR);
  	
    // Einzulesendes Joystick-Poti bestimmen
    ADMUX=(ADMUX & 0xE0) | (JOY_POTI_0+joy.potipointer);
  	
    // ADC einschalten, ADC-Vorteiler=128, AD-Wandlung starten und 
    // ADC-Interrupt einschalten
    ADCSRA=ADCSRA | (1<<ADEN) | (1<<ADPS2) | (1<<ADPS1) | 
										(1<<ADPS0) | (1<<ADSC) | (1<<ADIE);
  	
  }  // joy_init
  
  /* Initialisierung des externen EEPROM */
  void ext_eeprom_init (void)
  {
  
    // SCL und SDA als Eingang setzen
    EXT_EEPROM_DDR=EXT_EEPROM_DDR & ~((1<<EXT_EEPROM_SCL) | 
																			(1<<EXT_EEPROM_SDA));
    
    // keinen Vorteiler
    TWSR=0;
    
    // Berechnung der Bit-Rate
    TWBR=((F_CPU/SCL_CLOCK)-16)/2;
  	
  }  // ext_eeprom_init
  
  /* C Initialisieren */
  void uc_init (void)
  {
  
    // Unbenutzter Pin PC3: als Eingang mit Pull-Up definieren
		PORTC=PORTC | (1<<PC3);
		DDRC=DDRC & ~(1<<PC3);
		
		// Timer0-Init
    // keinen Vorteiler 
    TCCR0=(1<<CS00);
    
    // Timer2-Init
    // Vorteiler F_CPU/8
    TCCR2=(1<<CS21);
    
    // Werte fr Entprellung und Wiederholfunktion berechnen
    debounce=(DEBOUNCE_TIME*F_CPU)/(256*TIMER2PRESCALER);
    repeat_start=REPEAT_START_TIME/DEBOUNCE_TIME;
    repeat_next=REPEAT_NEXT_TIME/DEBOUNCE_TIME;
    
    // Timer-Werte initialisieren
    timer0count=0;
    timer2count=0;
  	
    // Timer-Interrupt Enable
    TIMSK=(1<<TOIE0) | (1<<TOIE2);
    sei();
  	
  }  // uc_init
  
  /* druckt die Zeichenkette (s[]) in Spalte (x) und Zeile (y) 
     auf die LCD-Anzeige */
  void lcd_print (unsigned char x,unsigned char y,char s[])
  {
  unsigned char a;
    
    // Zeile und Spalte bestimmen
    lcd_hex_8(0x80+(y-1)*0x40+x-1,0);
    wait_us(50);
  	
    // Text ausgeben
    for (a=0;s[a]!=0;a++)
    {
      lcd_hex_8(s[a],1);
      wait_us(50);  
    }  
  	
  }  // lcd_print
  
  /* Abfrage der 16er-Tastatur, Rckgabe der gedrckten Taste 
     0 = keine Taste gedrckt, 1 = links oben, 2 = links 2. von oben
  	 ... 16 = rechts unten */
  unsigned char key16_scan (void)
  {
  unsigned char column,line,key,keyin;
  
    key=0;
    
    for (line=0;line<=3 && key==0;line++)
    {
      // Zeile als Ausgang schalten und auf Low setzen
      KEY16_LINE_DDR=KEY16_LINE_DDR | (1<<(KEY16_LINE_1+line));
      KEY16_LINE_PORT=KEY16_LINE_PORT & ~(1<<(KEY16_LINE_1+line));
  		
      // kleine Pause
      asm("NOP");
      asm("NOP");
  		 
      // Spalte einlesen
      column=((KEY16_COLUMN_IN>>KEY16_COLUMN_1)&0xF);
  		
      // Zeile einlesen
      keyin=((KEY16_LINE_IN>>KEY16_LINE_1)&0xF);
  		 
      // Zeile wieder als Eingang setzten
      KEY16_LINE_DDR=KEY16_LINE_DDR & ~(1<<(KEY16_LINE_1+line));
      KEY16_LINE_PORT=KEY16_LINE_PORT | (1<<(KEY16_LINE_1+line));
  			
      // Ist in jeder Zeile und Spalte nur eine Taste gedrckt?
      // Mehrere Tasten diagonal werden nicht erkannt
      if (((column==0xE) || (column==0xD) || (column==0xB) || (column==0x7)) && 
						((keyin==0xE) || (keyin==0xD) || (keyin==0xB) || (keyin==0x7)))
      {
        // Berechnung der gedrckten Taste
        switch (column)
        {
          case 0xE : 
          {
  	    key=line+1;
  	    break;
          }
  	  case 0xD : 
          {
            key=line+5; 
            break;
          }
  	  case 0xB : 
          {
  	    key=line+9;
  	    break;
          }
  	  case 0x7 : 
          {
  	    key=line+13;
  	    break;
          }
        }  // switch (column)
  			
      }  // if (((column==0xE) || (column==0xD) || (column==0xB) || 
         //      (column==0x7)) && ((keyin==0xE) || (keyin==0xD) || 
         //      (keyin==0xB) || (keyin==0x7)))
  
    }  // for (line=0;line<=3 && key==0;line++)
    
    return key;
     	
  }  // key16_scan
  
  /* Startet bertragung mit externem EEPROM und sendet die HW-Adresse */
  void ext_eeprom_start (unsigned char address)
  {
  
    // sende Start-Bedingung
    TWCR=(1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
  
    // warte bis bertragung abgeschlossen ist
    while (!(TWCR & (1<<TWINT))) {}
  
    // sende HW-Adresse
    TWDR=address;
    TWCR=(1<<TWINT) | (1<<TWEN);
  
    // warte bis bertragung abgeschlossen ist
    while (!(TWCR & (1<<TWINT))) {}
  	
  }  // ext_eeprom_start
  
  /* Startet bertragung mit externem EEPROM und sendet die HW-Adresse 
     wartet dabei bis EEPROM ansprechbar ist */
  void ext_eeprom_start_wait (unsigned char address)
  {
  unsigned char twst;
  
    while (1)
    {
      // sende Start-Bedingung
      TWCR=(1<<TWINT) | (1<<TWSTA) | (1<<TWEN);
       
      // warte bis bertragung abgeschlossen ist
      while (!(TWCR & (1<<TWINT))) {}
       
      // bertragung zum EEPROM erfolgreich?
      twst=TW_STATUS & 0xF8;
      if ((twst!=TW_START) && (twst!=TW_REP_START))
      {
        continue;
      }
       
      // sende HW-Adresse
      TWDR=address;
      TWCR=(1<<TWINT) | (1<<TWEN);
       
      // warte bis bertragung abgeschlossen ist
      while (!(TWCR & (1<<TWINT))) {}
       
      // bertragung zum EEPROM erfolgreich?
      twst=TW_STATUS & 0xF8;
      if ((twst==TW_MT_SLA_NACK) || (twst==TW_MR_DATA_NACK)) 
      {    	    
        // EEPROM nicht ansprechbar, sende Stop-Bedingung
        TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
  			 
        // warte bis stop-Bedingung bertragen wurde
        while (TWCR & (1<<TWSTO)) {}
  			 
        continue;
      }  // if ((twst==TW_MT_SLA_NACK) || (twst==TW_MR_DATA_NACK))
      break;
    }
  	
  }  // ext_eeprom_start_wait
  
  /* Beendet die bertragung zum externen EEPROM */
  void ext_eeprom_stop (void)
  {
  
    // sende Stop-Bedingung
    TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
  	
    // warte bis stop-Bedingung bertragen wurde
    while (TWCR & (1<<TWSTO)) {}
  	
  }  // ext_eeprom_stop
  
  /* Sendet ein Byte zum externen EEPROM */
  void ext_eeprom_write_byte (unsigned char data)
  {	
  
    // sende data zum externen EEPROM
    TWDR=data;
    TWCR=(1<<TWINT) | (1<<TWEN);
  
    // warte bis bertragung abgeschlossen ist
    while (!(TWCR & (1<<TWINT))) {}
  	
  }  // ext_eeprom_write_byte
  
  /* Liest ein Byte vom externen EEPROM, weitere Bytes knnen gelesen werden */
  unsigned char ext_eeprom_read_ack (void)
  {
  
    TWCR=(1<<TWINT) | (1<<TWEN) | (1<<TWEA);
  	
    // warte bis bertragung abgeschlossen ist
    while (!(TWCR & (1<<TWINT))) {}
  
    return TWDR;
  }  // ext_eeprom_read_ack
  
  /* Liest ein letztes Byte vom externen EEPROM */
  unsigned char ext_eeprom_read_nak (void)
  {
  
    TWCR=(1<<TWINT) | (1<<TWEN);
  	
    // warte bis bertragung abgeschlossen ist
    while (!(TWCR & (1<<TWINT))) {}
  	
    return TWDR;
  	
  }  // ext_eeprom_read_nak
  
  /* Speichert (data[]) im externem EEPROM ab (address), letztes Byte ist 0 */
  void ext_eeprom_write_buffer (unsigned int address, unsigned char data[])
  {
  unsigned char i;
    
    i=0;
    while (data[i]!=0)
    {
      // EEPROM-Adresse senden
      ext_eeprom_start_wait(EXT_EEPROM_ADDRESS+EXT_EEPROM_WRITE);
  		
      // Zuerst High- und dann Low-Byte der Speicher-Adresse senden
			ext_eeprom_write_byte((address>>8) & 0xFF);
      ext_eeprom_write_byte(address & 0xFF);
  		
      // Daten senden bis EEPROM-Seitengrenze erreicht oder keine Daten mehr
      do
      {
        ext_eeprom_write_byte(data[i]);
        address++;
        i++;
      }
      while (address%128!=0 && data[i]!=0);
  		
      // Datenbertragung beenden
      ext_eeprom_stop();
    }  // while (data[i]!=0)
  	
    // EEPROM-Adresse senden
    ext_eeprom_start_wait(EXT_EEPROM_ADDRESS+EXT_EEPROM_WRITE);
  	
    // Zuerst High- und dann Low-Byte der Speicher-Adresse senden
    ext_eeprom_write_byte((address>>8) & 0xFF);
    ext_eeprom_write_byte(address & 0xFF);
  	
    // End-Markierung senden
    ext_eeprom_write_byte(0);
  	
    // Datenbertragung beenden
    ext_eeprom_stop();
  	
  }  // ext_eeprom_write_buffer
  
  /* Liest aus dem externem EEPROM ab (address) bis eine 0 oder FF gelesen 
     wird und speichert in (data[]) */
  void ext_eeprom_read_buffer (unsigned int address, unsigned char data[])
  {
  unsigned char i,d;
  	
    // EEPROM-Adresse senden
    ext_eeprom_start_wait(EXT_EEPROM_ADDRESS+EXT_EEPROM_WRITE);
    
    // Zuerst High- und dann Low-Byte der Speicher-Adresse senden
    ext_eeprom_write_byte((address>>8) & 0xFF);
    ext_eeprom_write_byte(address & 0xFF);
    
    // EEPROM auf Lesen stellen
    ext_eeprom_start(EXT_EEPROM_ADDRESS+EXT_EEPROM_READ);
    
    // Daten empfangen
    d=1;
    for (i=0;d!=0xFF && d!=0;i++)
    {
      d=ext_eeprom_read_ack();
      data[i]=d;
    }
    ext_eeprom_read_nak();
    
    // Datenbertragung beenden
    ext_eeprom_stop();
  	
  }  // ext_eeprom_read_buffer
  
  /* Sendet an die PC-Clockleitung einen Low-High-Impuls */
  void pc_lowhigh (void)
  {
  
    _delay_us(20);
  	
    // Clockleitung Low
    PS2_PORT=PS2_PORT & ~(1<<PC_CLOCK);
    PS2_DDR=PS2_DDR | (1<<PC_CLOCK);
    _delay_us(40);
  	
    // Clockleitung High
    PS2_DDR=PS2_DDR & ~(1<<PC_CLOCK);
    _delay_us(20);
  	
  }  // pc_lowhigh
  
  /* Sendet (sdata) an den PC */
  void ch_to_pc (unsigned char sdata)
  {
  unsigned char parity,datbit,in;
  
    // Timer2-Interrupt ausschalten
    TIMSK=TIMSK & ~(1<<TOIE2);
    
    // Warten bis Clock- und Datenleitung fr mindestens 50s high sind
    in=PS2_IN;
    timer0count=0;
    while (!(in & (1<<PC_CLOCK)) || !(in & (1<<PC_DATA)) || timer0count<5) 
    {
      in=PS2_IN;
      if (!(in & (1<<PC_CLOCK)) || !(in & (1<<PC_DATA)))
      {
        timer0count=0;
      }
    }  // while (!(in & (1<<PC_CLOCK)) || !(in & (1<<PC_DATA)) || timer0count<5)
  	
    parity=0;
  	
    // Startbit, Datenleitung Low
    PS2_PORT=PS2_PORT & ~(1<<PC_DATA);
    PS2_DDR=PS2_DDR | (1<<PC_DATA);
    pc_lowhigh();
   
    // 8 Datenbits senden
    for (datbit=0;datbit<=7;datbit++)
    {
      // wenn Bit = 1
      if (sdata & (1<<datbit))
      {
        PS2_DDR=PS2_DDR & ~(1<<PC_DATA);
        parity++;
      }
      // wenn Bit = 0
      else
      {
        PS2_PORT=PS2_PORT & ~(1<<PC_DATA);
        PS2_DDR=PS2_DDR | (1<<PC_DATA);
      }
      pc_lowhigh();
    }  // for (datbit=0;datbit<=7;datbit++)
   
    // ungerade Paritt senden
    // gerade Anzahl von Einsen
    if (!(parity % 2))
    {
      PS2_DDR=PS2_DDR & ~(1<<PC_DATA);
    }
    // ungerade Anzahl von Einsen
    else
    {
      PS2_PORT=PS2_PORT & ~(1<<PC_DATA);
      PS2_DDR=PS2_DDR | (1<<PC_DATA);
    }
    pc_lowhigh();
   
    //Stopbit senden
    PS2_DDR=PS2_DDR & ~(1<<PC_DATA);
    pc_lowhigh();
    
    // Timer2-Interrupt wieder einschalten
    TIMSK=TIMSK | (1<<TOIE2);
  	
  }  // ch_to_pc
  
  // Tasten-Sequenz an PC senden
  void send_to_pc (unsigned char digit)
  {
  unsigned char i;
  unsigned int startaddress;
  
    // EEPROM-Adresse bestimmen
    if (set_mode==SET_PASSWORD)
    {
      startaddress=PASSWORD_ADDRESS+MAX_KEY_DATA*(passwordset*10+digit);
    }
    if (set_mode==SET_KEY)
    {
      startaddress=MAX_KEY_DATA*(keyset*10+digit);
    }
    if (set_mode==SET_JOY)
    {
      startaddress=JOYSET_ADDRESS+MAX_KEY_DATA*(joyset*12+digit);
    }
  	
    // Tasten-Sequenz aus externem EEPROM lesen
    ext_eeprom_read_buffer(startaddress,keybuffer.data);
  	
    // Verbindung zwischen Tastatur und PC getrennt
    AT_KEY_2_PC_SWITCH_PORT=AT_KEY_2_PC_SWITCH_PORT & ~(1<<AT_KEY_2_PC_SWITCH);
    _delay_us(1);
  	
    do
    {
      // Tasten-Sequenz an PC senden
      for (i=0;keybuffer.data[i]!=0x00 && keybuffer.data[i]!=0xFF;i++)
      {
        ch_to_pc(keybuffer.data[i]);
      }
      wait_ms(30);
    }
    while ((set_mode!=SET_JOY && key16.longpress) || 
  	   (set_mode==SET_JOY && digit<=3 && joy.btn_longpress) || 
  	   (set_mode==SET_JOY && digit>=4 && joy.poti_longpress));
    // solange bis Taste bzw. Joystick nicht mehr bettigt wird
  	
    // Normal-Modus, Tastatur mit PC verbunden
    AT_KEY_2_PC_SWITCH_PORT=AT_KEY_2_PC_SWITCH_PORT | (1<<AT_KEY_2_PC_SWITCH);
    _delay_us(1);
  	
  }  // send_to_pc
  
  /* Anzeige des aktuellen Modus (Keyset, Joyset oder Passwordset) */
  void show_set (void)
  {
    switch (set_mode)
    {
      case SET_KEY :
      {
        sprintf(lcdstr,"Keyset: %u",keyset);
        break;
      }
      case SET_JOY :
      {
        sprintf(lcdstr,"Joyset: %u",joyset);
        break;
      }
      case SET_PASSWORD :
      {
        sprintf(lcdstr,"Passwordset: %u",passwordset);
        break;
      }
    }
		lcd_clear();
    lcd_print(1,1,lcdstr);
  	
  }  // show_set
  
  /* Bestimmen des Satzes (0-9) im aktuellen Modus
     (Keyset, Joyset oder Passwordset) */
  void set_keyset (unsigned char x,unsigned char y,unsigned char *set)
  {
  
    show_set();
    lcd_print(x,y,"?");
  	
    // Warten bis Ziffern-Taste gedrckt wurde oder Abbruch
    while (digit16[key16.key]==-1 && key16.key!=KEY_CANCEL) {}
    if (key16.key!=KEY_CANCEL)
    {
      *set=digit16[key16.key];
    }
    show_set();
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
  }  // set_keyset
  
  /* Eingabe der PIN */
  void enter_pin (unsigned char pin[])
  {
  unsigned char i;
  
    for (i=0;i<5 && key16.key!=KEY_CANCEL;i++)
    {
      // Warten bis Ziffern-Taste gedrckt wurde oder Abbruch
      while (digit16[key16.key]==-1 && key16.key!=KEY_CANCEL) {}
      if (key16.key!=KEY_CANCEL)
      {
        pin[i]=digit16[key16.key]+1;
        lcd_print(5+i,2,"*");
  			
        // Warten bis Taste losgelassen wurde
        while (key16.key!=0) {}
  			
      }  // if (key16.key!=KEY_CANCEL)
    }  // for (i=0;i<5 && key16.key!=KEY_CANCEL;i++)
  	
  }  // enter_pin
  
  /* Eintritt in den Passwort-Modus */
  void set_keypasswordset (void)
  {
  unsigned char i,pin[6],oldpin[6];
  
    // PIN aus EEPROM lesen
    ext_eeprom_read_buffer(PIN_ADDRESS,oldpin);
  	
    // wenn Modus ungleich Passwort-Modus und PIN vorhanden
    if (set_mode!=SET_PASSWORD && oldpin[0]!=0xFF)
    {
      lcd_clear();
      lcd_print(3,1,"Enter PIN!");
      lcd_print(5,2,"-----");
  		
      enter_pin(pin);
  		
      // PIN-Eingabe abgeschlossen
      if (key16.key!=KEY_CANCEL)
      {
        for (i=0;i<5 && oldpin[i]==pin[i];i++) {}
        // PIN nicht korrekt eingegeben
        if (i!=5)
        {
          lcd_clear();
          lcd_print(5,1,"Error!");
          wait_ms(MS2WAIT);
          show_set();
          return;
        }  // if (i!=5)
      }  // if (key16.key!=KEY_CANCEL)
  		
      // PIN-Eingabe wurde durch CANCEL abgebrochen
      else 
      {
        show_set();
        return;
      }
    }  // if (set_mode!=SET_PASSWORD && oldpin[0]!=0xFF)
  	
    //  Bestimmen des Passwort-Satzes
    set_mode=SET_PASSWORD;
    lcd_clear();
    lcd_print(1,1,"Passwordset:");
    set_keyset(14,1,&passwordset);
  	
  }  // set_keypasswordset
  
  /* LCD-Hintergrundbeleuchtung und -Kontrast verndern */
  void setup_backlight_contrast (unsigned char setupitem)
  {
  unsigned char exitmenu,oldvalue,buffer[2];
  unsigned int startaddress;
  
    exitmenu=NO;
    oldvalue=*ledvo_trig;
    lcd_clear();
    lcd_print(1,1,lcdstr);
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
    while (exitmenu==NO)
    {
      while (digit16[key16.key]==-1 && 
             key16.key!=KEY_OK && key16.key!=KEY_CANCEL) {}
  		
      // Wurde eine der Zifferntasten gedrckt?
      if (digit16[key16.key]!=-1)
      {
        // LCD-Hintergrundbeleuchtung bzw. -Kontrast verndern
        *ledvo_trig=digit16[key16.key];
        buffer[0]=*ledvo_trig;
  			
        // neuen Wert anzeigen
        sprintf(lcdstr,"%u",*ledvo_trig);
        lcd_print(16,1,lcdstr);
  			
        // Warten bis Taste losgelassen wurde
        while (key16.key!=0) {}
      }  // if (digit16[key16.key]!=-1)
  		
      // Abbruch
      if (key16.key==KEY_CANCEL)
      {
        // LCD-Hintergrundbeleuchtung bzw. -Kontrast wieder alten Wert zuweisen
        *ledvo_trig=oldvalue;
        exitmenu=CANCEL;
      }
  		
      // Neuen Wert speichern
      if (key16.key==KEY_OK)
      {
        if (oldvalue!=*ledvo_trig)
        {
          if (setupitem==SETUP_BACKLIGHT)
          {
            startaddress=BACKLIGHT_ADDRESS;
            sprintf(lcdstr,"Backlight saved!");
          }
          else
          {
            startaddress=CONTRAST_ADDRESS;
            sprintf(lcdstr,"Contrast saved!");
          }
  				
          // Neuen Wert in externem EEPROM speichern
          buffer[1]=0;
          ext_eeprom_write_buffer(startaddress,buffer);
          lcd_clear();
          lcd_print(1,1,lcdstr);
          wait_ms(MS2WAIT);
        }  // if (oldvalue!=*ledvo_trig)
        exitmenu=OK;
      }  // if (key16.key==KEY_OK)
    }  // while (exitmenu==NO)
  	
    // Setup wieder anzeigen
    lcd_clear();
    lcd_print(1,1,"Setup:");
    if (setupitem==SETUP_BACKLIGHT)
    {
      sprintf(lcdstr,"LCD-Backlight: %u",*ledvo_trig);
    }
    else
    {
      sprintf(lcdstr,"LCD-Contrast:  %u",*ledvo_trig);
    }
    lcd_print(1,2,lcdstr);
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
  }  // setup_backlight_contrast
  
  /* Zweifache Eingabe einer neuen PIN */
  void enter_new_pins (void)
  {
  unsigned char i,pin[6],oldpin[6];
  
    lcd_clear();
    lcd_print(2,1,"Enter New PIN!");
    lcd_print(5,2,"-----");
  	
    // PIN das erste mal eingeben
    enter_pin(oldpin);
  	
    // PIN-Eingabe abgeschlossen
    if (key16.key!=KEY_CANCEL)
    {
      lcd_clear();
      lcd_print(2,1,"Re-Enter PIN!");
      lcd_print(5,2,"-----");
  		
      // PIN das zweite mal zur Sicherheit eingeben
      enter_pin(pin);
  		
      // zweite PIN-Eingabe abgeschlossen
      if (key16.key!=KEY_CANCEL)
      {
        // zwei mal der gleiche PIN eingegeben
        for (i=0;i<5 && oldpin[i]==pin[i];i++) {}
        if (i==5)
        {
          pin[5]=0;
  				
          // neue PIN speichern
          ext_eeprom_write_buffer(PIN_ADDRESS,pin);
  				
          lcd_clear();
          lcd_print(4,1,"PIN saved!");
          wait_ms(MS2WAIT);
          return;
        }  // if (i==5)
  			
        // Fehler bei der zweiten Eingabe der PIN
        lcd_clear();
        lcd_print(5,1,"Error!");
        wait_ms(MS2WAIT);
  			
      }  // if (key16.key!=KEY_CANCEL)
    }  // if (key16.key!=KEY_CANCEL)
  	
  }  // enter_new_pins
  
  /* Neue PIN eingeben */
  void setup_pin (void)
  {
  unsigned char i,pin[6],oldpin[6];
  
    // alte PIN aus EEPROM lesen
    ext_eeprom_read_buffer(PIN_ADDRESS,oldpin);
  	
    // Es steht noch keine PIN im EEPROM
    if (oldpin[0]==0xFF)
    {
      enter_new_pins();
    }
  	
    // Es steht bereits eine PIN im EEPROM
    else
    {
      lcd_clear();
      lcd_print(2,1,"Enter Old PIN!");
      lcd_print(5,2,"-----");
  		
      enter_pin(pin);
  		
      // PIN-Eingabe abgeschlossen
      if (key16.key!=KEY_CANCEL)
      {
        // alte PIN korrekt eingegeben
        for (i=0;i<5 && oldpin[i]==pin[i];i++) {}
        if (i==5)
        {
          enter_new_pins();  
        }
  			
        // alte PIN nicht korrekt eingegeben
        else
        {
          lcd_clear();
          lcd_print(5,1,"Error!");
          wait_ms(MS2WAIT); 
        }
  			
      }  // if (key16.key!=KEY_CANCEL)
    }  // if (oldpin[0]!=0xFF)
  
    // Setup wieder anzeigen
    lcd_clear();
    lcd_print(1,1,"Setup:");
    lcd_print(1,2,"PIN");
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
   
  }  // setup_pin
  
  /* Joystick Kalibrieren */
  void setup_joystick (void)
  
  {
  struct joy_min_max
  {
    unsigned char min,mid,max;
  };
  
  unsigned char i,r_max,r_min,r_mid,joycalbuffer[9];
  struct joy_min_max joytotal[4];
  	
    lcd_clear();
    lcd_print(1,1,"Center Joystick");
    lcd_print(1,2,"and press button");
  	
    // Mittelstellung bestimmen
    while (joy.button[0]==1 && joy.button[1]==1 && joy.button[2]==1 && 
  	   joy.button[3]==1 && key16.key!=KEY_CANCEL) {}
  	
    for (i=0;i<=3;i++)
    {
      if (joy.poti[i]>20)
      {
        joytotal[i].min=joy.poti[i];
        joytotal[i].max=joy.poti[i];
        joytotal[i].mid=joy.poti[i];
      }
      else
      {
        joytotal[i].min=5;
        joytotal[i].max=5;
        joytotal[i].mid=5;
      }
      joy.cal[i].trig_mid_max=5;
      joy.cal[i].trig_mid_min=5;
    }  // for (i=0;i<=3;i++)
  	
    lcd_clear();
    lcd_print(2,1,"Move Joystick");
    lcd_print(1,2,"and press button");
  	
    // Warten bis kein Joystick-Knopf mehr gedrckt ist
    while (joy.button[0]==0 || joy.button[1]==0 || 
						joy.button[2]==0 || joy.button[3]==0) {}
  	
    // Minimale und Maximale Werte bestimmen
    while (joy.button[0]==1 && joy.button[1]==1 && joy.button[2]==1 && 
						joy.button[3]==1 && key16.key!=KEY_CANCEL)
    {
      for (i=0;i<=3;i++)
      {
        if (joy.poti[i]>20)
        {
          if (joy.poti[i]>joytotal[i].max)
          {
            joytotal[i].max=joy.poti[i];
          }
          if (joy.poti[i]<joytotal[i].min)
          {
            joytotal[i].min=joy.poti[i];
          }
        }  // if (joy.poti[i]>20)
      }  // for (i=0;i<=3;i++)
    }  // while (joy.button[0]==1 && joy.button[1]==1 && joy.button[2]==1 &&
			 //        joy.button[3]==1 && key16.key!=KEY_CANCEL)
  	
    // Warten bis kein Joystick-Knopf mehr gedrckt ist
    while (joy.button[0]==0 || joy.button[1]==0 || 
  	   joy.button[2]==0 || joy.button[3]==0) {}
  	
    if (key16.key!=KEY_CANCEL)
    {
      // Auslse-Werte bestimmen, bei denen eine Bewegung des Joysticks 
      // erkannt wird,
      // von der Mitte nach Auen bei 1/2 Vollausschlag
      for (i=0;i<=3;i++)
      {
        // Joystick-Poti angeschlossen?
        if (joytotal[i].mid>20)
        {
          // Zuerst Rckrechnung auf Widerstand, da AD-Werte nicht linear 
          // zum Widerstand sind
          r_min=(100*255/joytotal[i].max)-100;
          r_max=(100*255/joytotal[i].min)-100;
          r_mid=(100*255/joytotal[i].mid)-100;				
  				
          // Bestimmung der Auslsepunkte
          joy.cal[i].trig_mid_max=100*255/(r_min+((r_mid-r_min)/2)+100);
          joy.cal[i].trig_mid_min=100*255/(r_mid+((r_max-r_mid)/2)+100);
  				
        }  // if (joytotal[i].mid>20)
      }  // for (i=0;i<=3;i++)
  		
      // Anlegen des Feldes zum Speichern der Kalibrier-Werte
      for (i=0;i<=3;i++)
      {
        joycalbuffer[i*2]=joy.cal[i].trig_mid_max;
        joycalbuffer[i*2+1]=joy.cal[i].trig_mid_min;
      }
      joycalbuffer[8]=0;
  		
      // Kalibrier-Werte speichern
      ext_eeprom_write_buffer(JOYCAL_ADDRESS,joycalbuffer);
  		
      lcd_clear();
      lcd_print(4,1,"Joystick");
      lcd_print(3,2,"calibrated!");
      wait_ms(MS2WAIT);
  		
    }  // if (key16.key!=KEY_CANCEL)
  	
    // Setup wieder anzeigen
    lcd_clear();
    lcd_print(1,1,"Setup:");
    lcd_print(1,2,"Joystick");
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
  }  // setup_joystick
  
  // Bestimmen ob Joystick bettigt wurde (Rckgabe=1)
  unsigned char joypress (void)
  {
  unsigned char i;
  
    for (i=0;i<=3;i++)
    {
      if (joy.button[i]==0 || (joy.poti[i]>20 && 
         (joy.poti[i]<joy.cal[i].trig_mid_min || 
          joy.poti[i]>joy.cal[i].trig_mid_max)))
      {
        return 1;
      }
    }
    return 0;
  	
  }  // joypress
  
  /* Programmieren einer Tastatur-Sequenz auf eine Taste der 16er-Tastatur 
     bzw. einer Joystick-Funktion */
  void program_keyset (void)
  {
  unsigned char i,byteleft,savebyteleft,old,breakcode,extendet,atkey,
                joyfunc,oldjoyfunc,store;
  unsigned int startaddress;
  char joyfuncstr[][9]={"Button 1","Button 2","Button 3","Button 4",
													"Left 1","Up 1","Left 2","Up 2",
													"Right 1","Down 1","Right 2","Down 2"};
  
    // Variablen initialisieren
    for (i=0;i<MAX_KEY_DATA;i++)
    {
      keybuffer.data[i]=0;
    }
    keybuffer.pointer=0;
    byteleft=MAX_KEY_DATA-1;
    savebyteleft=byteleft;
    joyfunc=0;
    oldjoyfunc=1;
    old=0;
    breakcode=0;
    extendet=0;
    store=0;
    keyat.bitcount=11;
    keyat.finish=0;
  	
    // Verbindung zwischen Tastatur und PC getrennt
    AT_KEY_2_PC_SWITCH_PORT=AT_KEY_2_PC_SWITCH_PORT & ~(1<<AT_KEY_2_PC_SWITCH);
    _delay_us(1); 
  	
    // Setze Interrupt auf fallende Flanke
    keyat.clockedge=0;
    MCUCR=MCUCR & ~(1<<ISC00);
  	
    // PC-Tastatur-Interrupt einschalten
    GICR=GICR | (1<<INT0);
  	
    // Anzeige des richtigen Modus
    switch (set_mode)
    {
      case SET_KEY :
      {
        sprintf(lcdstr,"Program KS: %u",keyset);
        break;
      }
      case SET_JOY :
      {
        sprintf(lcdstr,"Program JS: %u",joyset);
        break;
      }
      case SET_PASSWORD :
      {
        sprintf(lcdstr,"Program PS: %u",passwordset);
        break;
      }
    }  // switch (set_mode)
    lcd_clear();
    lcd_print(1,1,lcdstr);
  	
    // Anzeige der freien Bytes
    sprintf(lcdstr,"Byte left: %u",byteleft);
    lcd_print(1,2,lcdstr);
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
    while (key16.key!=KEY_CANCEL && 
          ((set_mode!=SET_JOY && digit16[key16.key]==-1) || 
           (set_mode==SET_JOY && joypress()==0)))
    {
      // Anzeige der freien Bytes
      if (byteleft!=savebyteleft)
      {
        sprintf(lcdstr,"%u  ",byteleft);
        lcd_print(12,2,lcdstr);
        savebyteleft=byteleft;
      }
  		
      while (!keyat.finish && key16.key!=KEY_CANCEL && 
            ((set_mode!=SET_JOY && digit16[key16.key]==-1) || 
             (set_mode==SET_JOY && joypress()==0))) {}
  		
      // Neues Zeichen von der PS/2-Tastatur empfangen
      if (keyat.finish && byteleft>0)
      {
        atkey=keyat.data;
  			
        if (breakcode==2)
        {
          breakcode=0;
          old=0;
        }
        if (breakcode==1)
        {
          breakcode=2;
        }
        // Breakcode eingeleitet
        if (atkey==0xF0)
        {
          breakcode=1;
        }
  			
        if (extendet==2)
        {
          extendet=0;
          old=0;
        }
        if (extendet==1)
        {
          // Widerholfunktion bei Extendet-Keycode ausschalten
          if (keybuffer.pointer>2 && 
              keybuffer.data[keybuffer.pointer-3]==0xE0 && 
              keybuffer.data[keybuffer.pointer-2]==atkey)
          {
            keybuffer.pointer--;
            keybuffer.data[keybuffer.pointer]=0;
            byteleft++;
            old=atkey;  
          }
          extendet=2;
        }  // if (extendet==1)
  			
        // Extendet Keycode eingeleitet
        if (atkey==0xE0)
        {
          extendet=1;  
        }
  			
        // Neues empfangenes Zeichen ist ungleich dem vorigen Zeichen
        if (old!=atkey)
        {
          // Zeichen speichern
          keybuffer.data[keybuffer.pointer]=atkey;
          keybuffer.pointer++;
          old=atkey;
          byteleft--;
        }  // if (old!=atkey)
        keyat.finish=0;
  			
      }  // if (keyat.finish && byteleft>0)
    }  // while (key16.key!=KEY_CANCEL && 
       //       ((set_mode!=SET_JOY && digit16[key16.key]==-1) || 
       //        (set_mode==SET_JOY && joypress()==0)))
  	
    // Es wurden Zeichen auf der PS/2-Tastatur eingegeben
    if (keybuffer.pointer>0)
    {
      keybuffer.data[keybuffer.pointer]=0;
  		
      // Joystick-Modus
      if (set_mode==SET_JOY && key16.key!=KEY_CANCEL)
      {
        lcd_clear();
        lcd_print(1,1,"Show Joy-Func");
        lcd_print(13,2,"[OK]");
  			
        // Warten bis Taste losgelassen wurde
        while (key16.key!=0) {}
  			
        // Bestimmen der Joystick-Funktion
        while (key16.key!=KEY_OK && key16.key!=KEY_CANCEL)
        {
          // Joystick-Knopf
          for (i=0;i<=3;i++)
          {
            if (joy.button[i]==0)
            {
              joyfunc=i;
            }
          }  // for (i=0;i<=3;i++)
  				
          // Joystick_Achse
          if (joy.potiaxis>0)
          {
            joyfunc=joy.potiaxis+3;
          }
  				
          // Anzeigen der Joystick-Funktion
          if (oldjoyfunc!=joyfunc)
          {
            oldjoyfunc=joyfunc;
            lcd_print(1,2,"          ");
            lcd_print(1,2,joyfuncstr[joyfunc]);
          }
        }  // while (key16.key!=KEY_OK && key16.key!=KEY_CANCEL)
        if (key16.key==KEY_OK)
        {
          // Tasten-Sequenz im externen EEPROM speichern
          startaddress=JOYSET_ADDRESS+MAX_KEY_DATA*(joyset*12+joyfunc);
          store=1;
        }
      }  // if (set_mode==SET_JOY && key16.key!=KEY_CANCEL)
  		
      // Tastatur- bzw. Passwort-Modus
      else
      {
        if (digit16[key16.key]!=-1)
        {
          // Tasten-Sequenz im externen EEPROM speichern
          if (set_mode==SET_PASSWORD)
          {
            startaddress=PASSWORD_ADDRESS+
                          MAX_KEY_DATA*(passwordset*10+digit16[key16.key]);
          }
          else
          {
            startaddress=MAX_KEY_DATA*(keyset*10+digit16[key16.key]);
          }
          store=1;
        }  // if (digit16[key16.key]!=-1)
      }  // if !(set_mode==SET_JOY && key16.key!=KEY_CANCEL)
    }  // if (keybuffer.pointer>0)
  	
    if (store)
    {
      ext_eeprom_write_buffer(startaddress,keybuffer.data);
      lcd_clear();
      lcd_print(3,1,"Data saved!");
      wait_ms(MS2WAIT);
    }  // if (store)
  	
    // PC-Tastatur-Interrupt ausschalten
    GICR=GICR & ~(1<<INT0);
  	
    // Normal-Modus, Tastatur mit PC verbunden
    AT_KEY_2_PC_SWITCH_PORT=AT_KEY_2_PC_SWITCH_PORT | (1<<AT_KEY_2_PC_SWITCH);
    _delay_us(1);
  	
    // Setup wieder anzeigen
    lcd_clear();
    lcd_print(1,1,"Setup:");
    lcd_print(1,2,"Program");
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
  }  // program_keyset
  
  /* Wechselt den aktuellen Setup-Eintrag und zeigt ihn an */
  void setup_toggle (unsigned char *setupitem)
  {
  
    // Setup-Eintrag erhhen und bei berlauf auf Anfang setzen
    (*setupitem)++;
    if (*setupitem==LAST_SETUP_ITEM+1)
    {
      *setupitem=0;		    
    }
  	
    // Zeichenkette fr neuen Setup-Eintrag bestimmen
    switch (*setupitem)
    {
      case SETUP_PROGRAM :
      {
        sprintf(lcdstr,"Program         ");
        break;
      }
      case SETUP_JOYSTICK :
      {
        sprintf(lcdstr,"Joystick        ");
        break;
      }
      case SETUP_PIN :
      {
        sprintf(lcdstr,"PIN             ");
        break;
      }
      case SETUP_BACKLIGHT :
      {
        ledvo_trig=&lcd.led.trig;
        sprintf(lcdstr,"LCD-Backlight: %u",*ledvo_trig);
        break;
      }
      case SETUP_CONTRAST :
      {
        ledvo_trig=&lcd.vo.trig;
        sprintf(lcdstr,"LCD-Contrast:  %u",*ledvo_trig);
        break;
      }
    }  // switch (*setupitem)
  	
    // Neuen Setup-Eintrag anzeigen
    lcd_print(1,2,lcdstr);
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
  }  // setup_toggle
  
  /* Setup-Modus */
  void setup (void)
  {
  unsigned char setupitem;
  
    setupitem=0;
  	
    lcd_clear();
    lcd_print(1,1,"Setup:");
    lcd_print(1,2,"Program");
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
    while (key16.key!=KEY_CANCEL)
    {
      switch (key16.key)
      {
        // Setup-Eintrag wechseln
        case KEY_SETUP :
        {
          setup_toggle(&setupitem);
          break;
        }
  			
        // Setup-Eintrag auswhlen
        case KEY_OK :
        {
          switch (setupitem)
          {
            case SETUP_PROGRAM :
            {
              program_keyset();
              break;
            }
            case SETUP_JOYSTICK :
            {
              setup_joystick();
              break;
            }
            case SETUP_PIN :
            {
              setup_pin();
              break;
            }
            case SETUP_BACKLIGHT :
            case SETUP_CONTRAST :
            {
              setup_backlight_contrast(setupitem);
              break;
            }
  					
          }  //switch (setupitem)
        }  //case KEY_OK
  			
        break;
  			
      }  // switch (key16.key)
    }  // while (key16.key!=KEY_CANCEL)
  	
    show_set();
  	
    // Warten bis Taste losgelassen wurde
    while (key16.key!=0) {}
  	
  }  // setup
  
  /* Interrupt-Routine zum Einlesen der PS/2-Tastatur */
  SIGNAL (SIG_INTERRUPT0)
  {
  
    // Fallende Flanke
    if (!keyat.clockedge)
    {
      // Bits 1 bis 8 enthalten die Daten
      if ((keyat.bitcount<11) && (keyat.bitcount>2))
      {
        keyat.data=(keyat.data>>1);
        if (PS2_IN & (1<<AT_KEY_DATA))
        {
          keyat.data=keyat.data | 0x80;
        }
      }  // if ((keyat.bitcount<11) && (keyat.bitcount>2))
  		
      // reagiere das nchste mal auf die steigende Flanke
      MCUCR=MCUCR | (1<<ISC00);
      keyat.clockedge=1;
    }  // if (!keyat.clockedge)
  	
    // Steigende Flanke
    else
    {
      // reagiere das nchste mal auf die fallende Flanke
      MCUCR=MCUCR & ~(1<<ISC00);
      keyat.clockedge=0;
      keyat.bitcount--;
      if (keyat.bitcount==0)
      {
        keyat.bitcount=11;
        keyat.finish=1;
      }
    }  // if (keyat.clockedge)
  	
  }  // SIGNAL (SIG_INTERRUPT0)
  
  /* Interrupt-Routine zur Steuerung der LCD-Hintergrundbeleuchtung und 
     des -Kontrasts */
  SIGNAL (SIG_OVERFLOW0)
  {
  
    if (lcd.led.count==0)
    {
      LCD_PORT=LCD_PORT & ~(1<<LCD_LED);
    }
    lcd.led.count++;
    if (lcd.led.count==(lcd.led.trig+1))
    {
      LCD_PORT=LCD_PORT | (1<<LCD_LED);
      lcd.led.count=0;
    }
    
    if (lcd.vo.count==0)
    {
      LCD_PORT=LCD_PORT | (1<<LCD_VO);
    }
  	lcd.vo.count++;
    if (lcd.vo.count==(lcd.vo.trig+1))
    {
      LCD_PORT=LCD_PORT & ~(1<<LCD_VO);
      lcd.vo.count=0;
    }
  	
    timer0count++;
  	
  }  // SIGNAL (SIG_OVERFLOW0)
  
  /* Interrupt-Routine zur Abfrage der 16er-Tastatur und der Joystick-Knpfe */
  INTERRUPT (SIG_OVERFLOW2)
  {
  static unsigned char actval,i;
    
    timer2count++;
    
    // 156 = ca. 20ms 
    if (timer2count==debounce)
    {
      // 16er-Tastatur abfragen
      actval=key16_scan();
  		
      // Taste entprellt?
      if (actval==key16.oldkey)
      {
        // aktuelle Taste ungleich voriger Taste?
        if (actval!=key16.prevkey)
        {
          key16.repeat=repeat_start;
          key16.longpress=0;
        }
  			
        // aktuelle Taste gleich voriger Taste?
        else
        {
          key16.repeat--;
          if (key16.repeat==0)
          {
            key16.repeat=repeat_next;
            if (actval!=0)
            {
              // aktuelle Taste wurde lnger gedrckt
              key16.longpress=1;
            }
          }  // if (key16.repeat==0)
        }  // if (actval==key16.prevkey)
        key16.key=actval;
        key16.prevkey=actval;
  			
      }  // if (actval==key16.oldkey)
  		
      key16.oldkey=actval;
  			
      if (!key16.key)
      {
        actval=(JOY_IN>>JOY_BUTTON_0) & 0x0F;
  			
        // Joystick-Knpfe entprellt?
        if (actval==joy.oldbuttons)
        {
          // aktuelle Joystick-Knpfe ungleich vorigen Joystick-Knpfen?
          if (actval!=joy.prevbuttons)
          {
            joy.btn_repeat=repeat_start;
            joy.btn_longpress=0;
          }
  				
          // aktuelle Joystick-Knpfe gleich vorigen Joystick-Knpfen?
          else
          {
            joy.btn_repeat--;
            if (joy.btn_repeat==0)
            {
              joy.btn_repeat=repeat_next;
              if (actval!=0x0F)
              {
                // aktuelle Joystick-Knpfe wurden lnger gedrckt
                joy.btn_longpress=1;
              }
            }  // if (joy.btn_repeat==0)
          }  // if (actval==joy.prevbuttons)
          for (i=0;i<=3;i++)
          {
            joy.button[i]=(actval>>i) & 1;
          }
          joy.prevbuttons=actval;
  				
        }  // if (actval==joy.oldbuttons)
  			
        joy.oldbuttons=actval;
  			
        if (actval==0x0F)
        {
          for (i=0;i<=3 && actval==0x0F;i++)
          {
            if (joy.poti[i]>20 && (joy.poti[i]<joy.cal[i].trig_mid_min || 
                                   joy.poti[i]>joy.cal[i].trig_mid_max))
            {
              // Speichere die benutzte Joystick-Achse und Joystick-Richtung
              if (joy.poti[i]>joy.cal[i].trig_mid_max)
              {
                actval=i+1;
              }
              else
              {
                actval=i+5;
              }
            }  // if (joy.poti[i]>20 && (joy.poti[i]<joy.cal[i].trig_mid_min || 
               // 											 joy.poti[i]>joy.cal[i].trig_mid_max))
          }  // for (i=0;i<=3 && actval==0x0F;i++)
  				
          // Keine Joystick-Achse ber Schaltpunkt bewegt
          if (actval==0x0F)
          {
            actval=0;
            joy.poti_longpress=0;
            joy.potiaxis=0;
            joy.prevpotiaxis=0;
            joy.oldpotiaxis=0;
          }  // if (actval==0x0F)
  				
          // Joystick-Achse "entprellt" (eher fr Joypad) 
          if (actval==joy.oldpotiaxis)
          {
            // aktuelle Joystick-Achse ungleich voriger Joystick-Achse?
            if (actval!=joy.prevpotiaxis)
            {
              joy.poti_repeat=repeat_start;
              joy.poti_longpress=0;  
            }
  					
            // aktuelle Joystick-Achse gleich voriger Joystick-Achse?
            else
            {
              joy.poti_repeat--;
              if (joy.poti_repeat==0)
              {
                joy.poti_repeat=repeat_next;
                if (actval!=0)
                {
                  // aktuelle Joystick-Achse wurde lnger bettigt
                  joy.poti_longpress=1;
                }
              }  // if (joy.poti_repeat==0)
            }  // if (actval==joy.prevpotiaxis)
            joy.potiaxis=actval;
            joy.prevpotiaxis=actval;
  					
          }  // if (actval==joy.oldpotiaxis)
  				
          joy.oldpotiaxis=actval;
  				
        }  // if (actval==0x0F)
      }  // if (!key16.key)
  		
      timer2count=0;
  		
    }  // if (timer2count==debounce)
  	
  }  // INTERRUPT (SIG_OVERFLOW2)
  
  /* Interrupt-Routine zur Abfrage der Joystick-Achsen */
  SIGNAL (SIG_ADC)
  {
  
    // Ergebnis der AD-Wandlung speichern
    joy.poti[joy.potipointer]=ADCH;
  	
    // zu nchstem AD-Kanal schalten
    joy.potipointer++;
    if (joy.potipointer==4)
    {
      joy.potipointer=0;
    }
  	
    // Einzulesendes Joystick-Poti bestimmen
    ADMUX=(ADMUX & 0xE0) | (JOY_POTI_0+joy.potipointer);
  	
    // AD-Wandlung starten
    ADCSRA=ADCSRA | (1<<ADSC);
  	
  }  // SIGNAL (SIG_ADC)
  
  // Hauptprogramm
  int main (void)
  {
  unsigned char i,buffer[9],joyfunc;
  
    // Gert initialisieren
    lcd_init();
    ps2_init();
    key16_init();
    joy_init();
    ext_eeprom_init();
    uc_init();
  	
    // Wert fr LCD-Hintergrundbeleuchtung aus externem EEPROM lesen
    ext_eeprom_read_buffer(BACKLIGHT_ADDRESS,buffer);
    if (buffer[0]==0xFF)
    {
      buffer[0]=9;
    }
    lcd.led.trig=buffer[0];
  	
    // Wert fr LCD-Kontrast aus externem EEPROM lesen
    ext_eeprom_read_buffer(CONTRAST_ADDRESS,buffer);
    if (buffer[0]==0xFF)
    {
      buffer[0]=8;
    }
    lcd.vo.trig=buffer[0];
  	
    // Joystick-Kalibrierwerte aus externem EEPROM lesen
    ext_eeprom_read_buffer(JOYCAL_ADDRESS,buffer);
    for (i=0;i<=3;i++)
    {
      joy.cal[i].trig_mid_max=buffer[i*2];
      joy.cal[i].trig_mid_min=buffer[i*2+1];
    }
    
    set_mode=SET_KEY;
    keyset=0;
    joyset=0;
    passwordset=0;
  	
    lcd_print(3,1,"ProggyKeyJoy");
    lcd_print(5,2,"SW: 0.89");
    wait_ms(1000);
  	
    show_set();
  	
    while (1)
    {
      switch (key16.key)
      {
        // Bestimmen des Keyset
        case KEY_KEYSET :
        {
          set_mode=SET_KEY;
          set_keyset(9,1,&keyset);
          break;
        }
  			
        // Bestimmen des Joystickset
        case KEY_JOYSET :
        {
          set_mode=SET_JOY;
          set_keyset(9,1,&joyset);
          break;
        }
  			
        // Bestimmen des Passordset
        case KEY_PASSWORD :
        {
          set_keypasswordset();
          break;
        }
  			
        // Aufrufen des Setup
        case KEY_SETUP :
        {
          setup();
          break;
        }
      }  // switch (key16.key)
  		
      // Tasten-Sequenz aus Keyset oder Passwordset an PC senden
      if (set_mode!=SET_JOY)
      {
        // Ziffern-Taste auf 16er-Tastatur gedrckt?
        if (digit16[key16.key]!=-1)
        {
          // Tasten-Sequenz an PC senden
          send_to_pc(digit16[key16.key]);
  				
          // Warten bis Taste losgelassen wurde oder lnger gedrckt ist
          while (key16.key!=0 && !key16.longpress) {}
  				
        }  // if (digit16[key16.key]!=-1)
      }  // if (set_mode!=SET_JOY)
  		
      // Tasten-Sequenz aus Joystickset an PC senden
      else
      {
        // Bestimmen der Joystick-Funktion
        joyfunc=0xFF;
        for (i=0;i<=3 && joyfunc==0xFF;i++)
        {
          // Joystick-Knopf 
          if (joy.button[i]==0)
          {
            joyfunc=i;
          }
        }  // for (i=0;i<=3 && joyfunc==0xFF;i++)
        if (joyfunc==0xFF && joy.potiaxis>0)
        {
          joyfunc=joy.potiaxis+3;
        }
  			
        if (joyfunc!=0xFF)
        {
          // Tasten-Sequenz an PC senden
          send_to_pc(joyfunc);
          i--;
  				
          // Warten bis Joystick-Funktion (Knopf oder Poti) losgelassen wurde
          // oder lnger gedrckt ist
          while (joyfunc<=3 && joy.button[i]==0 && !joy.btn_longpress) {}
          while (joyfunc>=4 && joyfunc<=7 && joy.potiaxis!=0 && 
                 !joy.poti_longpress) {}
          while (joyfunc>=8 && joyfunc<=11 && joy.potiaxis!=0 && 
                 !joy.poti_longpress) {}
  				
        }  // if (joyfunc!=0xFF)
      }  // if (set_mode==SET_JOY)
  		
    }  // while (1)
  	
    return 1;
  	
  }  // main 
